"""Responsible for reading and cleaning input data sources."""
from collections import defaultdict
from datetime import datetime as dt
import getpass
import json
import logging
from typing import Any, Dict, List

from bs4 import BeautifulSoup
import pandas as pd
from pandera import check_input, check_output
from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait

from pricer import config as cfg
from pricer import io, schema, utils

logger = logging.getLogger(__name__)


def get_bb_item_page(driver: webdriver, item_id: int) -> Dict[Any, Any]:
    """Get Booty Bay json info for a given item_id."""
    driver.get(cfg.us["bb_selenium"]["BB_ITEMAPI"] + str(item_id))
    soup = BeautifulSoup(driver.page_source)
    text = soup.find("body").text
    if "captcha" in text:
        driver.get("https://www.bootybaygazette.com/#us/grobbulus-a/item/6049")
        input("User action required")

        # Redo
        driver.get(cfg.us["bb_selenium"]["BB_ITEMAPI"] + str(item_id))
        soup = BeautifulSoup(driver.page_source)
        text = soup.find("body").text
    return json.loads(text)


def start_driver() -> webdriver:
    """Spin up selenium driver for Booty Bay scraping."""
    account = cfg.secrets.get("account")
    password = cfg.secrets.get("password")
    if not password:
        password = getpass.getpass("Password:")
    try:
        driver = webdriver.Chrome(cfg.us["bb_selenium"]["CHROMEDRIVER_PATH"])
        driver.implicitly_wait(cfg.us["bb_selenium"]["PAGE_WAIT"])
        driver.get(cfg.us["bb_selenium"]["BB_BASEURL"])

        WebDriverWait(driver, 20).until(
            EC.element_to_be_clickable((By.CLASS_NAME, "battle-net"))
        ).click()

        driver.find_element_by_id("accountName").send_keys(account)
        driver.find_element_by_id("password").send_keys(password)

        WebDriverWait(driver, 20).until(
            EC.element_to_be_clickable((By.ID, "submit"))
        ).click()
    except Exception:
        raise SystemError("Error connecting to bb")

    input("Ready to continue after authentication...")
    return driver


def get_bb_data() -> None:
    """Reads Booty Bay web API data using selenium and blizzard login."""
    driver = start_driver()
    # Get item_ids for user specified items of interest
    user_items = cfg.ui.copy()
    user_items.pop("Empty Vial")
    user_items.pop("Leaded Vial")
    user_items.pop("Crystal Vial")

    item_ids = utils.get_item_ids()
    items_ids = {k: v for k, v in item_ids.items() if k in user_items}

    # Get bb data from API
    bb_data: Dict[str, Dict[Any, Any]] = defaultdict(dict)
    for item, item_id in items_ids.items():
        bb_data[item] = get_bb_item_page(driver, item_id)

    driver.close()
    io.writer(bb_data, "raw", "bb_data", "json")


def clean_bb_data() -> None:
    """Parses all Booty Bay item json into tabular formats."""
    item_data = io.reader("raw", "bb_data", "json")

    bb_fortnight: List = []
    bb_history: List = []
    bb_listings: List = []
    bb_alltime: List = []
    bb_deposit: Dict[str, int] = {}

    for item, data in item_data.items():

        bb_fortnight_data = pd.DataFrame(data["history"][0])
        bb_fortnight_data["snapshot"] = pd.to_datetime(
            bb_fortnight_data["snapshot"], unit="s"
        )
        bb_fortnight_data["item"] = item
        bb_fortnight.append(bb_fortnight_data)

        bb_history_data = pd.DataFrame(data["daily"])
        bb_history_data["item"] = item
        bb_history.append(bb_history_data)

        if data["auctions"]["data"]:
            bb_listings_data = pd.DataFrame(data["auctions"]["data"])
            bb_listings_data = bb_listings_data[
                ["quantity", "buy", "sellerrealm", "sellername"]
            ]
            bb_listings_data["price_per"] = (
                bb_listings_data["buy"] / bb_listings_data["quantity"]
            ).astype(int)
            bb_listings_data = bb_listings_data.drop("sellerrealm", axis=1)
            bb_listings_data["item"] = item
            bb_listings.append(bb_listings_data)

        bb_alltime_data = pd.DataFrame(data["monthly"][0])
        bb_alltime_data["item"] = item
        bb_alltime.append(bb_alltime_data)

        vendorprice = item_data[item]["stats"][0]["selltovendor"]
        bb_deposit[item] = int(vendorprice / 20 * 12)

    bb_fortnight_df = pd.concat(bb_fortnight)
    bb_fortnight_df["snapshot"] = pd.to_datetime(bb_fortnight_df["snapshot"])

    bb_history_df = pd.concat(bb_history)
    bb_history_df["date"] = pd.to_datetime(bb_history_df["date"])

    bb_alltime_df = pd.concat(bb_alltime)
    bb_alltime_df["date"] = pd.to_datetime(bb_alltime_df["date"])

    bb_listings_df = pd.concat(bb_listings)
    bb_listings_df = bb_listings_df[bb_listings_df["price_per"] > 0]

    bb_deposit_df = pd.DataFrame.from_dict(bb_deposit, orient="index")
    bb_deposit_df.columns = ["deposit"]
    bb_deposit_df.index.name = "item"

    io.writer(bb_fortnight_df, "cleaned", "bb_fortnight", "parquet")
    io.writer(bb_history_df, "cleaned", "bb_history", "parquet")
    io.writer(bb_alltime_df, "cleaned", "bb_alltime", "parquet")
    io.writer(bb_listings_df, "cleaned", "bb_listings", "parquet")
    io.writer(bb_deposit_df, "cleaned", "bb_deposit", "parquet")


def get_arkinventory_data() -> None:
    """Reads WoW Addon Ark Inventory lua data and saves local copy as json."""
    acc_inv: dict = {}
    for account_name in cfg.us.get("accounts"):
        path = utils.make_lua_path(account_name, "ArkInventory")
        data = io.reader(name=path, ftype="lua")
        acc_inv = utils.source_merge(acc_inv, data).copy()

    arkinventory_data = acc_inv["ARKINVDB"]["global"]["player"]["data"]
    io.writer(arkinventory_data, "raw", "arkinventory_data", "json")


def clean_arkinventory_data(run_dt: dt) -> None:
    """Reads Ark Inventory json and parses into tabular format."""
    inventory_data = io.reader("raw", "arkinventory_data", "json")

    settings = io.reader("config", "general_settings", "yaml")

    raw_data: list = []
    monies: Dict[str, int] = {}
    for character, character_data in inventory_data.items():
        character_name = character.split(" ")[0]

        character_money = int(character_data.get("info").get("money", 0))
        monies[character] = character_money

        # Get Bank, Inventory, Character, Mailbox etc
        location_slots = character_data.get("location", [])

        for lkey in location_slots:
            items: Dict[str, int] = defaultdict(int)
            if str(lkey) not in settings["location_info"]:
                continue
            else:
                loc_name = settings["location_info"][str(lkey)]

            location_slot = location_slots[lkey]
            if location_slot:
                bag_slots = location_slot["bag"]

                # Get the items from each of the bags, add to master list
                for bag in bag_slots:
                    for item in bag.get("slot", []):
                        if item.get("h") and item.get("count"):
                            item_name = item.get("h").split("[")[1].split("]")[0]
                            items[item_name] += item.get("count")

            for item_name, item_count in items.items():
                raw_data.append((character_name, loc_name, item_name, item_count))

    # Convert information to dataframe
    cols = ["character", "location", "item", "count"]
    ark_inventory = pd.DataFrame(raw_data)
    ark_inventory.columns = cols
    ark_inventory["timestamp"] = run_dt
    io.writer(ark_inventory, "cleaned", "ark_inventory", "parquet")

    ark_monies = pd.Series(monies)
    ark_monies.name = "monies"
    ark_monies = pd.DataFrame(ark_monies)
    ark_monies["timestamp"] = run_dt
    io.writer(ark_monies, "cleaned", "ark_monies", "parquet")


def get_beancounter_data() -> None:
    """Reads WoW Addon Beancounter lua and saves to local json."""
    """Reads Ark Inventory json and parses into tabular format."""
    beancounter_data: dict = {}
    for account_name in cfg.us.get("accounts"):
        path = utils.make_lua_path(account_name, "BeanCounter")
        bean = io.reader(name=path, ftype="lua")
        beancounter_data = utils.source_merge(beancounter_data, bean).copy()
    io.writer(beancounter_data, "raw", "beancounter_data", "json")


def clean_beancounter_data() -> None:
    """Reads Beancounter json and parses into tabular format."""
    data = io.reader("raw", "beancounter_data", "json")

    item_names = {v: k for k, v in utils.get_item_ids().items()}

    # Parses all listings into flat python list
    parsed = []
    for character, auction_data in data["BeanCounterDB"]["Grobbulus"].items():
        for auction_type, item_listings in auction_data.items():
            for item_id, listings in item_listings.items():
                for _, listing in listings.items():
                    for auction in listing:
                        parsed.append(
                            [auction_type]
                            + [item_names[int(item_id)]]
                            + [character]
                            + auction.split(";")
                        )

    # Setup as pandas dataframe, remove irrelevant columns
    df = pd.DataFrame(parsed)

    bean_purchases = clean_beancounter_purchases(df)
    io.writer(bean_purchases, "cleaned", "bean_purchases", "parquet")

    failed = clean_beancounter_failed(df)
    success = clean_beancounter_success(df)

    bean_results = success.append(failed)
    bean_results["success"] = bean_results["auction_type"].replace(
        {"completedAuctions": 1, "failedAuctions": 0}
    )
    io.writer(bean_results, "cleaned", "bean_results", "parquet")


@check_input(schema.beancounter_data_raw_schema)
@check_output(schema.beancounter_purchases_schema)
def clean_beancounter_purchases(df: pd.DataFrame) -> pd.DataFrame:
    """Further processing of purchase beancounter data."""
    purchases = df[df[0] == "completedBidsBuyouts"]

    columns = [
        "auction_type",
        "item",
        "buyer",
        "qty",
        "drop_4",
        "drop_5",
        "drop_6",
        "buyout",
        "bid",
        "seller",
        "timestamp",
        "drop_11",
        "drop_12",
    ]
    purchases.columns = columns
    purchases = purchases.drop([col for col in columns if "drop_" in col], axis=1)

    purchases["qty"] = purchases["qty"].astype(int)
    purchases["buyout"] = purchases["buyout"].astype(float)
    purchases["bid"] = purchases["bid"].astype(int)

    purchases["buyout_per"] = purchases["buyout"] / purchases["qty"]
    purchases["bid_per"] = purchases["bid"] / purchases["qty"]

    purchases["timestamp"] = pd.to_datetime(purchases["timestamp"], unit="s")
    return purchases


@check_input(schema.beancounter_data_raw_schema)
@check_output(schema.beancounter_posted_schema)
def clean_beancounter_posted(df: pd.DataFrame) -> pd.DataFrame:
    """Further processing of posted auction beancounter data."""
    posted = df[df[0] == "postedAuctions"]

    columns = [
        "auction_type",
        "item",
        "seller",
        "qty",
        "buyout",
        "bid",
        "duration",
        "deposit",
        "timestamp",
        "drop_9",
        "drop_10",
        "drop_11",
        "drop_12",
    ]
    posted.columns = columns
    posted = posted.drop([col for col in columns if "drop_" in col], axis=1)

    posted["qty"] = posted["qty"].astype(int)
    posted["buyout"] = posted["buyout"].astype(float)
    posted["bid"] = posted["bid"].astype(int)
    posted["duration"] = posted["duration"].astype(int)
    posted["deposit"] = posted["deposit"].replace("", 0).astype(int)

    posted["buyout_per"] = posted["buyout"] / posted["qty"]
    posted["bid_per"] = posted["bid"] / posted["qty"]

    posted["timestamp"] = pd.to_datetime(posted["timestamp"], unit="s")
    return posted


@check_input(schema.beancounter_data_raw_schema)
@check_output(schema.beancounter_failed_schema)
def clean_beancounter_failed(df: pd.DataFrame) -> pd.DataFrame:
    """Further processing of failed auction beancounter data."""
    failed = df[df[0] == "failedAuctions"]

    columns = [
        "auction_type",
        "item",
        "seller",
        "qty",
        "drop_4",
        "deposit",
        "drop_6",
        "buyout",
        "bid",
        "drop_9",
        "timestamp",
        "drop_11",
        "drop_12",
    ]
    failed.columns = columns
    failed = failed.drop([col for col in columns if "drop_" in col], axis=1)

    col = ["qty", "deposit", "buyout", "bid"]
    failed[col] = failed[col].replace("", 0).astype(int)

    failed["buyout_per"] = failed["buyout"] / failed["qty"]
    failed["bid_per"] = failed["bid"] / failed["qty"]

    failed["timestamp"] = pd.to_datetime(failed["timestamp"], unit="s")
    return failed


@check_input(schema.beancounter_data_raw_schema)
@check_output(schema.beancounter_success_schema)
def clean_beancounter_success(df: pd.DataFrame) -> pd.DataFrame:
    """Further processing of successful auction beancounter data."""
    success = df[df[0] == "completedAuctions"]

    columns = [
        "auction_type",
        "item",
        "seller",
        "qty",
        "received",
        "deposit",
        "ah_cut",
        "buyout",
        "bid",
        "buyer",
        "timestamp",
        "drop_11",
        "drop_12",
    ]
    success.columns = columns
    success = success.drop([col for col in columns if "drop_" in col], axis=1)

    col = ["qty", "received", "deposit", "ah_cut", "buyout", "bid"]
    success[col] = success[col].replace("", 0).astype(int)

    success["received_per"] = success["received"] / success["qty"]
    success["buyout_per"] = success["buyout"] / success["qty"]
    success["bid_per"] = success["bid"] / success["qty"]

    success["timestamp"] = pd.to_datetime(success["timestamp"], unit="s")
    return success


def get_auctioneer_data() -> None:
    """Reads WoW Addon Auctioneer lua and parses text file into json."""
    ahm_account = [r["account"] for r in cfg.us.get("roles") if r.get("role") == "ahm"]
    path = utils.make_lua_path(ahm_account[0], "Auc-ScanData")
    ropes = io.reader(name=path, ftype="lua", custom="Auc-ScanData")

    listings = []
    for rope in ropes:
        if len(rope) < 10:
            continue
        listings_part = rope.split("},{")
        listings_part[0] = listings_part[0].split("{{")[1]
        listings_part[-1] = listings_part[-1].split("},}")[0]
        listings.extend(listings_part)
    aucscan_data = [x.split("|")[-1].split(",") for x in listings]

    io.writer(aucscan_data, "raw", "aucscan_data", "json")


@check_input(schema.auc_listings_raw_schema)
@check_output(schema.auc_listings_schema)
def process_auctioneer_data(df: pd.DataFrame) -> pd.DataFrame:
    """Performs processing of auctioneer data."""
    auction_timing: Dict[int, int] = {1: 30, 2: 60 * 2, 3: 60 * 12, 4: 60 * 24}

    df["time_remaining"] = df[6].astype(int).replace(auction_timing)
    df["item"] = df[8].str.replace('"', "").str[1:-1]
    df["quantity"] = df[10].replace("nil", 0).astype(int)
    df["buy"] = df[16].astype(int)
    df["sellername"] = df[19].str.replace('"', "").str[1:-1]

    df = df[df["quantity"] > 0]

    df["price_per"] = (df["buy"] / df["quantity"]).astype(int)
    df = df[df["price_per"] > 0]

    cols = ["item", "quantity", "buy", "sellername", "price_per", "time_remaining"]
    df = df[cols]
    return df


def clean_auctioneer_data() -> None:
    """Cleans Auctioneer json data into tablular format."""
    aucscan_data = io.reader("raw", "aucscan_data", "json")

    auc_listings_raw = pd.DataFrame(aucscan_data)
    auc_listings = process_auctioneer_data(auc_listings_raw)

    # Saves latest scan to intermediate (immediate)
    io.writer(auc_listings, "cleaned", "auc_listings", "parquet")


@check_input(schema.item_skeleton_raw_schema)
@check_output(schema.item_skeleton_schema)
def process_item_skeleton(df: pd.DataFrame) -> pd.DataFrame:
    """Make transformation to item skeleton."""
    df["made_from"] = df["made_from"] == df["made_from"]

    int_cols = ["min_holding", "max_holding", "vendor_price"]
    df[int_cols] = df[int_cols].fillna(0).astype(int)

    # df["max_sell"] = df["max_sell"].fillna(df["max_holding"]).astype(int)

    df["std_holding"] = (df["max_holding"] - df["min_holding"]) / 7
    df["mean_holding"] = df[["min_holding", "max_holding"]].mean(axis=1).astype(int)

    bool_cols = ["Buy", "Sell", "make_pass"]
    df[bool_cols] = df[bool_cols].fillna(False).astype(int)
    return df


def create_item_skeleton() -> None:
    """Creates basic dataframe from user items information."""
    user_items = cfg.ui.copy()
    item_skeleton_raw = pd.DataFrame(user_items).T

    item_skeleton = process_item_skeleton(item_skeleton_raw)

    io.writer(item_skeleton, "intermediate", "item_skeleton", "parquet")



"""Produces reporting to help interpret analysis and campaigns."""
import logging

from pricer import io

logger = logging.getLogger(__name__)


def have_in_bag() -> None:
    """Prints expected profits, make sure its in your bag."""
    sell_policy = io.reader("outputs", "sell_policy", "parquet")
    sell_policy = sell_policy.set_index("item")

    sell_policy = sell_policy[
        sell_policy["estimated_profit"] > sell_policy["feasible_profit"]
    ]

    print(sell_policy["estimated_profit"])



"""Contains helper functions to support data pipeline."""
import logging
from typing import Any, Dict, Tuple

from pricer import config as cfg, io

logger = logging.getLogger(__name__)


def get_seconds_played(time_played: str) -> int:
    """Convert string representation of time played to seconds."""
    try:
        days, hours, mins, seconds = time_played.split("-")
    except ValueError as error:
        logger.error(error)
        raise ValueError(
            "Play time not formatted correctly; needs to be '00d-00h-00m-00s'"
        )

    total_seconds = (
        int(days[:-1]) * 24 * 60 * 60
        + int(hours[:-1]) * 60 * 60
        + int(mins[:-1]) * 60
        + int(seconds[:-1])
    )

    return total_seconds


def duration_str_to_mins(dur_char: str = "m") -> int:
    """Convert duration string to auction minutes."""
    choices: Dict[str, int] = {"s": 120, "m": 480, "l": 1440}
    return choices[dur_char]


def source_merge(a: dict, b: dict, path: list = None) -> Dict[Any, Any]:
    """Merges b into a."""
    if path is None:
        path = []
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                source_merge(a[key], b[key], path + [str(key)])
            elif a[key] == b[key]:
                pass  # same leaf value
            else:
                pass  # raise Exception("Conflict at %s" % ".".join(path + [str(key)]))
        else:
            a[key] = b[key]
    return a


def make_lua_path(account_name: str = "", datasource: str = "") -> str:
    """Forms a path to a lua file."""
    warcraft_path = cfg.us.get("warcraft_path").rstrip("/")
    path = (
        f"{warcraft_path}/WTF/Account/{account_name}/" + f"SavedVariables/{datasource}"
    )
    return path


def get_item_ids() -> Dict[str, int]:
    """Read item id database."""
    item_codes = io.reader("static", "items", "csv")
    return item_codes.set_index("name")["entry"].to_dict()


def dict_to_lua(data: dict) -> str:
    """Converts python dict into long str."""
    lua_print = "\n"
    for key in data.keys():
        lua_print += f"{key} = " + dump_lua(data[key]) + "\n"
    return lua_print


def dump_lua(data: Any) -> Any:
    """Borrowed code to write python dict as lua format(ish)."""
    if type(data) is str:
        return f'"{data}"'
    if type(data) in (int, float):
        return f"{data}"
    if type(data) is bool:
        return data and "true" or "false"
    if type(data) is list:
        list_work = "{"
        list_work += ", ".join([dump_lua(item) for item in data])
        list_work += "}"
        return list_work
    if type(data) is dict:
        dict_work = "{"
        dict_work += ", ".join(
            [
                f"[{k}]={dump_lua(v)}" if type(k) is int else f'["{k}"]={dump_lua(v)}'
                for k, v in data.items()
            ]
        )
        dict_work += "}"
        return dict_work
    logger.warning(f"Lua parsing error; unknown type {type(data)}")


def find_attribute_location(content: bytes, initial_key: bytes) -> Tuple[int, int]:
    """Search binary lua for an attribute start and end location."""
    start = content.index(initial_key)

    brack = 0
    bracked = False
    for _end, char in enumerate(content[start:].decode("ascii")):
        if char == "{":
            brack += 1
            bracked = True
        if char == "}":
            brack -= 1
            bracked = True

        if brack == 0 and bracked:
            break
    _end += start + 1
    return start, _end



"""Pricer performs data analysis on the WoW auction house."""
from importlib_metadata import PackageNotFoundError, version

try:
    __version__ = version(__name__)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"



"""File read and writes."""
import json
import logging
from pathlib import Path
from typing import Any, BinaryIO

import pandas as pd
from slpp import slpp as lua
import yaml

from pricer import config as cfg, utils

logger = logging.getLogger(__name__)
pd.options.mode.chained_assignment = None  # default='warn'


def reader(schema: str = "", name: str = "", ftype: str = "", custom: str = "") -> Any:
    """Standard program writer, allows pathing extensibility i.e. testing or S3."""
    if ftype == "yaml":
        base_path = ""
    else:
        base_path = cfg.env["basepath"]
    path = Path(base_path, schema, name + "." + ftype)
    logger.debug(f"Reading {name} {ftype} from {path}")
    if ftype == "parquet":
        data = pd.read_parquet(path)
    elif ftype == "csv":
        data = pd.read_csv(path)
    elif ftype == "json":
        with open(path, "r") as json_r:
            data = json.load(json_r)
    elif ftype == "lua":
        if custom == "Auc-ScanData":
            data = []
            with open(path, "r") as lua_auc:
                on = False
                for line in lua_auc.readlines():
                    if on and "return" in line:
                        data.append(line)
                    elif '["ropes"]' in line:
                        on = True
        elif custom == "rb":
            with open(path, "rb") as lua_rb:  # type: BinaryIO
                data = lua_rb.read()
        else:
            with open(path, "r") as lua_r:
                data = lua.decode("{" + lua_r.read() + "}")
    elif ftype == "yaml":
        with open(path, "r") as yaml_r:
            data = yaml.safe_load(yaml_r)

    return data


def writer(
    data: Any, schema: str = "", name: str = "", ftype: str = "", custom: str = ""
) -> None:
    """Standard program writer, allows pathing extensibility i.e. testing or S3."""
    path = Path(cfg.env["basepath"], schema, name + "." + ftype)
    logger.debug(f"Writing {name} {ftype} to {path}")
    if ftype == "parquet":
        data.to_parquet(path, compression="gzip")
    elif ftype == "json":
        with open(path, "w") as json_w:
            json.dump(data, json_w, indent=4)
    elif ftype == "lua":
        if custom == "wb":
            with open(path, "wb") as lua_wb:  # type: BinaryIO
                lua_wb.write(data)
        else:
            with open(path, "w") as lua_w:
                lua_w.write(utils.dict_to_lua(data))



"""Collates information to form buy/sell campaigns."""
import logging
from typing import Any, Dict

import pandas as pd
from scipy.stats import norm
from slpp import slpp as lua

from pricer import config as cfg, io, utils

logger = logging.getLogger(__name__)


def analyse_buy_policy(MAX_BUY_STD: int = 2) -> None:
    """Create buy policy."""
    logger.debug(f"max buy std {MAX_BUY_STD}")

    item_table = io.reader("intermediate", "item_table", "parquet")

    buy_policy = item_table[item_table["Buy"] == True]
    subset_cols = [
        "pred_price",
        "pred_std",
        "inv_total_all",
        "replenish_qty",
        "std_holding",
        "replenish_z",
    ]
    buy_policy = buy_policy[subset_cols]

    listing_each = io.reader("intermediate", "listing_each", "parquet")

    listing_each = listing_each.sort_values("price_per")

    rank_list = listing_each.join(buy_policy, on="item").dropna()

    rank_list["rank"] = rank_list.groupby("item")["price_per"].rank(method="max")

    rank_list = rank_list.drop_duplicates()
    rank_list["updated_rank"] = rank_list["replenish_qty"] - rank_list["rank"]
    rank_list["updated_replenish_z"] = (
        rank_list["updated_rank"] / rank_list["std_holding"]
    )

    rank_list["updated_replenish_z"] = rank_list["updated_replenish_z"].clip(
        upper=MAX_BUY_STD
    )

    rank_list = rank_list[rank_list["updated_replenish_z"] > rank_list["pred_z"]]
    io.writer(rank_list, "reporting", "buy_rank", "parquet")

    buy_policy["buy_price"] = rank_list.groupby("item")["price_per"].max()
    buy_policy["buy_price"] = buy_policy["buy_price"].fillna(1).astype(int)

    buy_policy.index.name = "item"
    buy_policy = buy_policy.reset_index()
    io.writer(buy_policy, "outputs", "buy_policy", "parquet")


def encode_buy_campaign(buy_policy: pd.DataFrame) -> Dict[str, Dict[str, Any]]:
    """Encodes buy campaign dataframe into dictionary."""
    cols = ["item", "buy_price"]
    assert (buy_policy.columns == cols).all(), "Buy policy incorrectly formatted"
    buy_policy = buy_policy.set_index("item")

    item_ids = utils.get_item_ids()

    new_snatch = {}
    for item, b in buy_policy.iterrows():
        item_id = item_ids.get(item, None)
        if item_id:
            snatch_item: Dict[str, Any] = {}
            snatch_item["price"] = int(b["buy_price"])
            link_text = f"|cffffffff|Hitem:{item_id}::::::::39:::::::|h[{item}]|h|r"
            snatch_item["link"] = link_text
            new_snatch[f"{item_id}:0:0"] = snatch_item

    return new_snatch


def write_buy_policy() -> None:
    """Writes the buy policy to all accounts."""
    buy_policy = io.reader("outputs", "buy_policy", "parquet")

    cols = ["item", "buy_price"]
    new_snatch = encode_buy_campaign(buy_policy[cols])

    # Read client lua, replace with

    for account in cfg.us.get("accounts", []):
        path = utils.make_lua_path(account_name=account, datasource="Auc-Advanced")
        data = io.reader(name=path, ftype="lua")
        snatch = data["AucAdvancedData"]["UtilSearchUiData"]["Current"]
        snatch["snatch.itemsList"] = {}
        snatch = snatch["snatch.itemsList"]
        data["AucAdvancedData"]["UtilSearchUiData"]["Current"][
            "snatch.itemsList"
        ] = new_snatch
        io.writer(data, name=path, ftype="lua")


def encode_sell_campaign(sell_policy: pd.DataFrame) -> Dict[str, Any]:
    """Encode sell policy dataframe into dictionary."""
    cols = ["item", "proposed_buy", "proposed_bid", "sell_count", "stack", "duration"]
    assert (sell_policy.columns == cols).all(), "Sell policy incorrectly formatted"
    sell_policy = sell_policy.set_index("item")

    item_ids = utils.get_item_ids()

    # Seed new appraiser
    new_appraiser: Dict[str, Any] = {
        "bid.markdown": 0,
        "columnsortcurDir": 1,
        "columnsortcurSort": 6,
        "duration": 720,
        "bid.deposit": True,
    }

    for item, d in sell_policy.iterrows():
        code = item_ids[item]

        try:
            new_appraiser[f"item.{code}.fixed.bid"] = int(d["proposed_bid"])
        except ValueError:
            raise ValueError(f"{code} for {item} not present")

        new_appraiser[f"item.{code}.fixed.buy"] = int(d["proposed_buy"])
        new_appraiser[f"item.{code}.duration"] = int(d["duration"])
        new_appraiser[f"item.{code}.number"] = int(d["sell_count"])
        new_appraiser[f"item.{code}.stack"] = int(d["stack"])

        new_appraiser[f"item.{code}.bulk"] = True
        new_appraiser[f"item.{code}.match"] = False
        new_appraiser[f"item.{code}.model"] = "fixed"

    return new_appraiser


def write_sell_policy() -> None:
    """Writes the sell policy to accounts."""
    sell_policy = io.reader("outputs", "sell_policy", "parquet")

    cols = ["item", "proposed_buy", "proposed_bid", "sell_count", "stack", "duration"]
    new_appraiser = encode_sell_campaign(sell_policy[cols])

    # Read client lua, replace with
    for account in cfg.us.get("accounts", []):
        path = utils.make_lua_path(account_name=account, datasource="Auc-Advanced")
        data = io.reader(name=path, ftype="lua")
        data["AucAdvancedConfig"]["profile.Default"]["util"][
            "appraiser"
        ] = new_appraiser
        io.writer(data, name=path, ftype="lua")


def analyse_sell_policy(
    stack: int = 1,
    max_sell: int = 10,
    duration: str = "m",
    MAX_STD: int = 5,
    MIN_PROFIT: int = 300,
    MIN_PROFIT_PCT: float = 0.015,
) -> None:
    """Creates sell policy based on information."""
    item_table = io.reader("intermediate", "item_table", "parquet")
    listing_each = io.reader("intermediate", "listing_each", "parquet")
    item_volume_change_probability = io.reader(
        "intermediate", "item_volume_change_probability", "parquet"
    )

    cols = [
        "deposit",
        "material_costs",
        "pred_std",
        "pred_price",
        "max_sell",
        "inv_ahm_bag",
        "replenish_qty",
        "replenish_z",
    ]
    sell_items = item_table[item_table["Sell"] == True][cols]
    sell_items["deposit"] = sell_items["deposit"] * (
        utils.duration_str_to_mins(duration) / (60 * 24)
    )

    sell_items["exponential_percent"] = 2 - sell_items["replenish_z"].apply(
        lambda x: norm.cdf(x)
    )

    listing_each = listing_each[listing_each["pred_z"] < MAX_STD]
    listing_each = listing_each.sort_values(["item", "price_per"])
    listing_each["rank"] = (
        listing_each.groupby("item")["pred_z"].rank(method="first").astype(int) - 1
    )

    listing_each = pd.merge(
        item_volume_change_probability, listing_each, how="left", on=["item", "rank"]
    )
    listing_each = listing_each.set_index(["item"])
    listing_each["pred_z"] = listing_each["pred_z"].fillna(MAX_STD)

    gouge_price = sell_items["pred_price"] + (sell_items["pred_std"] * MAX_STD)

    listing_each["price_per"] = (
        listing_each["price_per"].fillna(gouge_price).astype(int)
    )
    listing_each = listing_each.reset_index().sort_values(["item", "rank"])

    listing_profits = pd.merge(
        listing_each, sell_items, how="left", left_on="item", right_index=True
    )

    listing_profits["proposed_buy"] = listing_profits["price_per"] - 9

    listing_profits["estimated_profit"] = (
        (listing_profits["proposed_buy"] * 0.95 - listing_profits["material_costs"])
        * (listing_profits["probability"] ** listing_profits["exponential_percent"])
    ) - (listing_profits["deposit"] * (1 - listing_profits["probability"]))

    best_profits_ind = listing_profits.groupby("item")["estimated_profit"].idxmax()
    sell_policy = listing_profits.loc[best_profits_ind]

    sell_policy["min_profit"] = MIN_PROFIT
    sell_policy["profit_pct"] = MIN_PROFIT_PCT * sell_policy["pred_price"]
    sell_policy["feasible_profit"] = sell_policy[["min_profit", "profit_pct"]].max(
        axis=1
    )
    sell_policy["infeasible"] = (
        sell_policy["feasible_profit"] > sell_policy["estimated_profit"]
    )

    # Shows the amount required to be profitable
    sell_policy["proposed_bid"] = (
        sell_policy["proposed_buy"]
        - sell_policy["estimated_profit"]
        + sell_policy["feasible_profit"]
    )

    low_bid_ind = sell_policy[
        sell_policy["proposed_bid"] < sell_policy["proposed_buy"]
    ].index
    sell_policy.loc[low_bid_ind, "proposed_bid"] = sell_policy.loc[
        low_bid_ind, "proposed_buy"
    ]

    sell_policy["duration"] = utils.duration_str_to_mins(duration)
    sell_policy = sell_policy.sort_values("estimated_profit", ascending=False)

    sell_policy["stack"] = stack
    sell_policy["max_sell"] = sell_policy["max_sell"].replace(0, max_sell)
    sell_policy["sell_count"] = sell_policy[["inv_ahm_bag", "max_sell"]].min(axis=1)
    sell_policy["sell_count"] = (
        sell_policy["sell_count"] / sell_policy["stack"]
    ).astype(int)

    sell_policy["min_sell"] = sell_policy[["max_sell", "inv_ahm_bag"]].min(axis=1)
    adjust_stack = sell_policy[sell_policy["min_sell"] < sell_policy["stack"]].index
    sell_policy.loc[adjust_stack, "stack"] = 1
    sell_policy.loc[adjust_stack, "sell_count"] = sell_policy.loc[
        adjust_stack, "min_sell"
    ]

    io.writer(sell_policy, "outputs", "sell_policy", "parquet")

    listing_profits = listing_profits.set_index(["rank", "item"])[
        "estimated_profit"
    ].unstack()
    io.writer(listing_profits, "reporting", "listing_profits", "parquet")


def analyse_make_policy() -> None:
    """Prints what potions to make."""
    item_table = io.reader("intermediate", "item_table", "parquet")
    item_table.index.name = "item"

    cols = [
        "item_id",
        "made_from",
        "make_pass",
        "inv_total_all",
        "mean_holding",
        "inv_ahm_bag",
        "inv_ahm_bank",
        "Sell",
    ]
    make = item_table[cols]

    user_items = cfg.ui.copy()

    make["make_ideal"] = make["mean_holding"] - make["inv_total_all"]
    make["make_counter"] = make["make_ideal"].apply(lambda x: max(x, 0))
    make["make_mat_available"] = make["inv_ahm_bag"] + make["inv_ahm_bank"]
    make["make_actual"] = 0
    make["make_mat_flag"] = 0

    # Iterates through the table one at a time, to ensure fair distribution of mat usage
    # Tests if reached counter and is made from stuff
    # Checks the material count can go down first before decrementing
    # If after each check, append to list to see for any changes on any pass through
    change = [True]
    while any(change):
        change = []

        for item, row in make.iterrows():

            made_from = user_items[item].get("made_from", {})
            under_counter = row["make_actual"] < row["make_counter"]
            make_pass = row["make_pass"]

            if made_from and under_counter and not (make_pass):
                item_increment = True
                for material, qty in made_from.items():
                    if "Vial" not in material:
                        item_increment = (
                            make.loc[material, "make_mat_available"] >= qty
                        ) & item_increment

                if item_increment:
                    for material, qty in user_items[item].get("made_from", {}).items():
                        make.loc[material, "make_mat_available"] -= qty
                        make.loc[material, "make_mat_flag"] = 1
                    make.loc[item, "make_actual"] += 1

                change.append(item_increment)

    make_me = make[(make["make_pass"] == 0) & (make["make_actual"] > 0)]["make_actual"]
    make_me.name = "Automake"
    print(make_me)

    make_main = make[(make["make_pass"] == 1) & (make["make_ideal"] > 0)]["make_ideal"]
    make_main.name = "Make on main"
    print(make_main)

    make_should = make[
        (
            (make["make_pass"] == 0)
            & (make["make_ideal"] > make["make_actual"])
            & (make["made_from"] == 1)
        )
    ]["make_ideal"]
    make_should.name = "Missing mats"
    print(make_should)

    materials_list = make[make["make_mat_flag"] == 1]["item_id"]
    sell_items = make[make["Sell"] == 1]["item_id"]

    path = utils.make_lua_path(
        account_name="396255466#1", datasource="TradeSkillMaster"
    )
    content = io.reader(name=path, ftype="lua", custom="rb")

    start, end = utils.find_attribute_location(
        content, b'["f@Alliance - Grobbulus@internalData@crafts"]'
    )

    crafting = content[start:end]
    crafting_dict = lua.decode("{" + crafting.decode("ascii") + "}")

    for _, item_data in crafting_dict[
        "f@Alliance - Grobbulus@internalData@crafts"
    ].items():
        item_name = item_data.get("name")
        if item_name in make_me:
            queued = int(make_me.loc[item_name])
        else:
            queued = 0

        if "queued" in item_data:
            item_data["queued"] = queued

    new_crafting = utils.dict_to_lua(crafting_dict).encode("ascii")
    new_crafting = new_crafting.replace(
        b"\nf@Alliance - Grobbulus@internalData@crafts",
        b'\n["f@Alliance - Grobbulus@internalData@crafts"]',
    )

    content = content[:start] + new_crafting + content[end:]

    start, end = utils.find_attribute_location(content, b'["p@Default@userData@items"]')

    item_text = '["p@Default@userData@items"] = {'
    for _, item_code in materials_list.iteritems():
        item_text += f'["i:{item_code}"] = "Herbs", '

    for item_name, item_code in sell_items.items():
        if item_name not in materials_list:
            item_text += f'["i:{item_code}"] = "Sell", '

    item_text += "}"

    content = content[:start] + item_text.encode("ascii") + content[end:]
    io.writer(content, name=path, ftype="lua", custom="wb")
    io.writer(make, "outputs", "make", "parquet")



"""Main entry point for pricer program."""
import argparse
from datetime import datetime as dt
import logging
import warnings

from . import analysis, campaign, config as cfg, reporting, sources

warnings.simplefilter(action="ignore")
logger = logging.getLogger(__name__)


def main() -> None:
    """Main program runner."""
    run_dt = dt.now().replace(microsecond=0)

    parser = argparse.ArgumentParser(description="WoW Auctions")
    parser.add_argument("-b", help="Update web booty bay analysis", action="store_true")

    parser.add_argument("-s", type=int, default=5, help="Stack size")
    parser.add_argument("-m", type=int, default=20, help="Max sell")
    parser.add_argument("-d", type=str, default="m", help="Duration")

    parser.add_argument("-v", help="Verbose mode (info)", action="store_true")
    parser.add_argument("-vv", help="Verbose mode (debug)", action="store_true")
    args = parser.parse_args()

    cfg.set_loggers(base_logger=logger, v=args.v, vv=args.vv)
    logger.info("Program started, arguments parsed")
    logger.debug(args)

    if args.b:
        sources.get_bb_data()

    sources.clean_bb_data()
    sources.get_arkinventory_data()
    sources.clean_arkinventory_data(run_dt)
    sources.get_beancounter_data()
    sources.clean_beancounter_data()
    sources.get_auctioneer_data()
    sources.clean_auctioneer_data()
    sources.create_item_skeleton()

    analysis.predict_item_prices()
    analysis.analyse_listing_minprice()
    analysis.analyse_material_cost()
    analysis.create_item_inventory()
    analysis.analyse_listings()
    analysis.analyse_replenishment()

    analysis.create_item_table()

    analysis.predict_volume_sell_probability(args.d)

    campaign.analyse_buy_policy()
    campaign.write_buy_policy()

    campaign.analyse_sell_policy(stack=args.s, max_sell=args.m, duration=args.d)
    campaign.write_sell_policy()

    campaign.analyse_make_policy()

    reporting.have_in_bag()

    logger.info(f"Program end, seconds {(dt.now() - run_dt).total_seconds()}")


if __name__ == "__main__":
    main()



"""Analyses cleaned data sources to form intermediate tables."""
import logging
from typing import List

from numpy import inf
import pandas as pd
from scipy.stats import gaussian_kde

from pricer import config as cfg, io, utils

logger = logging.getLogger(__name__)


def predict_item_prices(quantile: float = 0.025) -> None:
    """Analyse exponential average mean and std of items given 14 day, 2 hour history."""
    bb_fortnight = io.reader("cleaned", "bb_fortnight", "parquet")

    user_items = cfg.ui.copy()

    # Work out if an item is auctionable, or get default price
    item_prices = pd.DataFrame()
    for item_name, item_details in user_items.items():
        vendor_price = item_details.get("vendor_price")

        if vendor_price:
            item_prices[item_name] = vendor_price
        else:
            df = bb_fortnight[bb_fortnight["item"] == item_name]
            df["silver"] = df["silver"].clip(
                lower=df["silver"].quantile(quantile),
                upper=df["silver"].quantile(1 - quantile),
            )
            try:
                item_prices.loc[item_name, "pred_price"] = int(
                    df["silver"].ewm(alpha=0.2).mean().iloc[-1]
                )
                item_prices.loc[item_name, "pred_std"] = df["silver"].std().astype(int)
            except IndexError:
                logging.exception(
                    f"""Price prediction problem for {item_name}.
                    Did you add something and not use booty bay?"""
                )

    qty_df = bb_fortnight[bb_fortnight["snapshot"] == bb_fortnight["snapshot"].max()]
    qty_df = qty_df.set_index("item")["quantity"]
    qty_df.name = "pred_quantity"

    predicted_prices = item_prices.join(qty_df).fillna(0).astype(int)
    io.writer(predicted_prices, "intermediate", "predicted_prices", "parquet")


def analyse_listing_minprice() -> None:
    """Determine current Auctions minimum price. TODO Is this needed?"""
    auc_listings = io.reader("cleaned", "auc_listings", "parquet")

    # Note this SHOULD be a simple groupby min, but getting 0's for some strange reason!
    item_mins = {}
    for item in auc_listings["item"].unique():
        x = auc_listings[(auc_listings["item"] == item)]
        item_mins[item] = int(x["price_per"].min())

    listings_minprice = pd.DataFrame(pd.Series(item_mins)).reset_index()
    listings_minprice.columns = ["item", "listing_minprice"]
    io.writer(listings_minprice, "intermediate", "listings_minprice", "parquet")


def analyse_material_cost() -> None:
    """Analyse cost of materials for items, using purchase history or BB predicted price."""
    bean_purchases = io.reader("cleaned", "bean_purchases", "parquet")
    item_skeleton = io.reader("intermediate", "item_skeleton", "parquet")
    item_prices = io.reader("intermediate", "predicted_prices", "parquet")

    user_items = cfg.ui.copy()
    user_buys = [k for k, v in user_items.items() if v.get("Buy")]

    bean_purchases = bean_purchases[bean_purchases["item"].isin(user_buys)].sort_values(
        ["item", "timestamp"]
    )

    item: List = sum(
        bean_purchases.apply(lambda x: [x["item"]] * x["qty"], axis=1).tolist(), []
    )
    price_per: List = sum(
        bean_purchases.apply(lambda x: [x["buyout_per"]] * x["qty"], axis=1).tolist(),
        [],
    )
    purchase_each = pd.DataFrame([item, price_per], index=["item", "price_per"]).T

    ewm = (
        purchase_each.groupby("item")
        .apply(lambda x: x["price_per"].ewm(span=100).mean())
        .reset_index()
    )
    purchase_rolling = purchase_each.join(ewm, rsuffix="_rolling")
    purchase_rolling = purchase_rolling.drop_duplicates("item", keep="last")
    purchase_rolling = purchase_rolling.set_index("item")["price_per_rolling"].astype(
        int
    )

    item_skeleton.index.name = "item"
    mat_prices = item_skeleton.join(purchase_rolling).join(item_prices)

    mat_prices["material_price"] = (
        mat_prices["price_per_rolling"]
        .fillna(mat_prices["pred_price"])
        .fillna(mat_prices["vendor_price"])
        .astype(int)
    )

    user_items = cfg.ui.copy()

    # Determine raw material cost for manufactured items
    item_costs = {}
    for item_name, item_details in user_items.items():
        material_cost = 0
        made_from = item_details.get("made_from", {})
        if made_from:
            for ingredient, count in made_from.items():
                material_cost += mat_prices.loc[ingredient, "material_price"] * count
        else:
            material_cost = mat_prices.loc[item_name, "material_price"]
        item_costs[item_name] = int(material_cost)

    material_costs = pd.DataFrame.from_dict(item_costs, orient="index").reset_index()
    material_costs.columns = ["item", "material_costs"]
    io.writer(material_costs, "intermediate", "material_costs", "parquet")


def create_item_inventory() -> None:
    """Convert Arkinventory tabular data into dataframe of counts for user items."""
    item_inventory = io.reader("cleaned", "ark_inventory", "parquet")

    roles = {char["name"]: char["role"] for char in cfg.us.get("roles", {})}

    item_inventory["role"] = item_inventory["character"].apply(
        lambda x: roles[x] if x in roles else "char"
    )
    role_types = ["ahm", "mule", "char"]
    assert item_inventory["role"].isin(role_types).all()

    location_rename = {
        "Inventory": "bag",
        "Bank": "bank",
        "Auctions": "auc",
        "Mailbox": "mail",
    }
    item_inventory["loc_short"] = item_inventory["location"].replace(location_rename)
    item_inventory["inv"] = (
        "inv_" + item_inventory["role"] + "_" + item_inventory["loc_short"]
    )

    item_inventory = item_inventory.groupby(["inv", "item"]).sum()["count"].unstack().T

    # Ensure 9x grid of columns
    for role in role_types:
        for loc in location_rename.values():
            col = f"inv_{role}_{loc}"
            if col not in item_inventory.columns:
                item_inventory[col] = 0

    item_inventory = item_inventory.fillna(0).astype(int)

    # Analyse aggregate; ordering important here
    item_inventory["inv_total_all"] = item_inventory.sum(axis=1)

    cols = [x for x in item_inventory.columns if "ahm" in x or "mule" in x]
    item_inventory["inv_total_hold"] = item_inventory[cols].sum(axis=1)

    cols = [x for x in item_inventory.columns if "ahm" in x]
    item_inventory["inv_total_ahm"] = item_inventory[cols].sum(axis=1)

    io.writer(item_inventory, "intermediate", "item_inventory", "parquet")


def analyse_listings() -> None:
    """Convert live listings into single items."""
    auc_listings = io.reader("cleaned", "auc_listings", "parquet")
    predicted_prices = io.reader("intermediate", "predicted_prices", "parquet")

    user_items = cfg.ui.copy()
    auc_listings = auc_listings[auc_listings["item"].isin(user_items)]

    ranges = pd.merge(
        auc_listings,
        predicted_prices,
        how="left",
        left_on="item",
        right_index=True,
        validate="m:1",
    )

    ranges["pred_z"] = (ranges["price_per"] - ranges["pred_price"]) / ranges["pred_std"]

    item: List = sum(
        ranges.apply(lambda x: [x["item"]] * x["quantity"], axis=1).tolist(), []
    )
    price_per: List = sum(
        ranges.apply(lambda x: [x["price_per"]] * x["quantity"], axis=1).tolist(), []
    )
    z: List = sum(
        ranges.apply(lambda x: [x["pred_z"]] * x["quantity"], axis=1).tolist(), []
    )

    listing_each = pd.DataFrame(
        [item, price_per, z], index=["item", "price_per", "pred_z"]
    ).T

    io.writer(listing_each, "intermediate", "listing_each", "parquet")


def analyse_replenishment() -> None:
    """Determine the demand for item replenishment."""
    item_skeleton = io.reader("intermediate", "item_skeleton", "parquet")
    item_inventory = io.reader("intermediate", "item_inventory", "parquet")

    item_skeleton.index.name = "item"
    replenish = item_skeleton.join(item_inventory).fillna(0).astype(int)

    user_items = cfg.ui.copy()

    replenish["replenish_qty"] = replenish["mean_holding"] - replenish["inv_total_all"]

    # Update replenish list with made_from
    for item, row in replenish.iterrows():
        if row["replenish_qty"] > 0:
            for ingredient, count in user_items[item].get("made_from", {}).items():
                replenish.loc[ingredient, "replenish_qty"] += (
                    count * row["replenish_qty"]
                )

    replenish["replenish_z"] = replenish["replenish_qty"] / replenish["std_holding"]
    replenish["replenish_z"] = (
        replenish["replenish_z"].replace([inf, -inf], 0).fillna(0)
    )

    replenish = replenish[["replenish_qty", "replenish_z"]].reset_index()
    io.writer(replenish, "intermediate", "replenish", "parquet")


def create_item_table() -> None:
    """Combine item information into single master table."""
    item_skeleton = io.reader("intermediate", "item_skeleton", "parquet")
    material_costs = io.reader("intermediate", "material_costs", "parquet")
    bb_deposit = io.reader("cleaned", "bb_deposit", "parquet")
    listings_minprice = io.reader("intermediate", "listings_minprice", "parquet")
    item_inventory = io.reader("intermediate", "item_inventory", "parquet")
    predicted_prices = io.reader("intermediate", "predicted_prices", "parquet")
    replenish = io.reader("intermediate", "replenish", "parquet")

    item_table = (
        item_skeleton.join(material_costs.set_index("item"))
        .join(listings_minprice.set_index("item"))
        .join(predicted_prices)
        .join(item_inventory)
        .join(bb_deposit)
        .fillna(0)
        .astype(int)
        .join(replenish.set_index("item"))
    )

    item_ids = utils.get_item_ids()

    item_table["item_id"] = item_table.index
    item_table["item_id"] = item_table["item_id"].apply(
        lambda x: item_ids[x] if x in item_ids else 0
    )

    io.writer(item_table, "intermediate", "item_table", "parquet")


def predict_volume_sell_probability(
    dur_char: str = "m", MAX_LISTINGS: int = 1000
) -> None:
    """Expected volume changes as a probability of sale given BB recent history."""
    bb_fortnight = io.reader("cleaned", "bb_fortnight", "parquet")

    user_sells = [k for k, v in cfg.ui.items() if v.get("Sell")]

    duration_mins = utils.duration_str_to_mins(dur_char)
    polls = int(duration_mins / 60 / 2)
    logger.debug(f"{polls} polls")

    item_volume_change_probability = pd.DataFrame(columns=user_sells)
    for item in user_sells:
        item_fortnight = bb_fortnight[bb_fortnight["item"] == item]

        volume_df = pd.DataFrame()
        price_df = pd.DataFrame()
        for i in range(1, polls + 1):
            item_fortnight["snapshot_prev"] = item_fortnight["snapshot"].shift(i)
            offset_test = pd.merge(
                item_fortnight,
                item_fortnight,
                left_on=["snapshot", "item"],
                right_on=["snapshot_prev", "item"],
            )
            volume_df[i] = offset_test["quantity_y"] - offset_test["quantity_x"]
            price_df[i] = offset_test["silver_y"] - offset_test["silver_x"]

        volume_df = volume_df.dropna().mean(axis=1)
        price_df = price_df.dropna().mean(axis=1)
        volume_df = volume_df[price_df <= 0]
        try:
            gkde = gaussian_kde(volume_df)
        except ValueError as e:
            raise ValueError(
                f"Could not analyse {item}, is this new and needs bb?"
            ) from e

        listing_range = range(-MAX_LISTINGS + 1, 1)
        probability = pd.Series(gkde(listing_range), index=listing_range)

        probability = probability.cumsum()
        probability.index = [-i for i in probability.index]

        item_volume_change_probability[item] = probability.sort_index()

    item_volume_change_probability.index.name = "rank"
    item_volume_change_probability.columns.name = "item"
    item_volume_change_probability = item_volume_change_probability.stack()
    item_volume_change_probability.name = "probability"
    item_volume_change_probability = item_volume_change_probability.reset_index()

    io.writer(
        item_volume_change_probability,
        "intermediate",
        "item_volume_change_probability",
        "parquet",
    )



"""It collates and loads user specified configuration for data pipeline."""
import logging
from typing import Any, Dict

from pricer import io

logger = logging.getLogger(__name__)


def set_loggers(
    base_logger: Any = None, v: bool = False, vv: bool = False, test: bool = False
) -> None:
    """Sets up logging across modules in project.

    Uses arguments originally specified from command line to set
    streaming logging level. Log files are written in debug mode.

    Args:
        base_logger: base logger object instantiated at program run time.
            Used to find other log objects across modules.
        v: Run in verbose mode (INFO)
        vv: Run in very verbose mode (DEBUG)
        test: Override / runs in verbose mode (INFO), in case not
            user specified.
    """
    if vv:
        log_level = 10
    elif v or test:
        log_level = 20
    else:
        log_level = 30

    loggers = [base_logger] + [
        logging.getLogger(name)  # type: ignore
        for name in logging.root.manager.loggerDict  # type: ignore
        if "pricer." in name
    ]
    for logger in loggers:
        logger.setLevel(log_level)  # type: ignore
        formatter = logging.Formatter("%(asctime)s:%(name)s:%(message)s")

        file_handler = logging.FileHandler(f"logs/{logger.name}.log")  # type: ignore
        file_handler.setLevel(10)
        file_handler.setFormatter(formatter)

        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(formatter)

        logger.addHandler(file_handler)  # type: ignore
        logger.addHandler(stream_handler)  # type: ignore


us = io.reader("config", "user_settings", "yaml")
ui = io.reader("config", "user_items", "yaml")
try:
    secrets = io.reader(name="SECRETS", ftype="yaml")
except FileNotFoundError:
    secrets = {"username": None, "password": None}

env: Dict[str, str] = {"basepath": "data"}



"""Schema enforcement for project."""
import pandera as pa
from pandera import Column, Index

item_skeleton_raw_schema = pa.DataFrameSchema(
    columns={
        "min_holding": Column(pa.Int, nullable=True),
        "max_holding": Column(pa.Int, nullable=True),
        "max_sell": Column(pa.Int, nullable=True),
        "Buy": Column(nullable=True),
        "Sell": Column(nullable=True),
        "made_from": Column(pa.Object, nullable=True),
        "make_pass": Column(pa.Int, nullable=True),
        "vendor_price": Column(pa.Int, nullable=True),
    },
    strict=True,
    index=Index(pa.String),
)

item_skeleton_schema = pa.DataFrameSchema(
    columns={
        "min_holding": Column(pa.Int),
        "max_holding": Column(pa.Int),
        "max_sell": Column(pa.Int, nullable=True),
        "Buy": Column(pa.Int),
        "Sell": Column(pa.Int),
        "made_from": Column(pa.Bool),
        "make_pass": Column(pa.Int),
        "vendor_price": Column(pa.Int),
        "std_holding": Column(pa.Float),
        "mean_holding": Column(pa.Int),
    },
    strict=True,
    index=Index(pa.String),
)

auc_listings_raw_schema = pa.DataFrameSchema(
    columns={
        0: Column(pa.String, nullable=True),
        1: Column(pa.String, nullable=True),
        2: Column(pa.String, nullable=True),
        3: Column(pa.String, nullable=True),
        4: Column(pa.String, nullable=True),
        5: Column(pa.String, nullable=True),
        6: Column(pa.String, nullable=True),
        7: Column(pa.String, nullable=True),
        8: Column(pa.String, nullable=True),
        9: Column(pa.String, nullable=True),
        10: Column(pa.String, nullable=True),
        11: Column(pa.String, nullable=True),
        12: Column(pa.String, nullable=True),
        13: Column(pa.String, nullable=True),
        14: Column(pa.String, nullable=True),
        15: Column(pa.String, nullable=True),
        16: Column(pa.String, nullable=True),
        17: Column(pa.String, nullable=True),
        18: Column(pa.String, nullable=True),
        19: Column(pa.String, nullable=True),
        20: Column(pa.String, nullable=True),
        21: Column(pa.String, nullable=True),
        22: Column(pa.String, nullable=True),
        23: Column(pa.String, nullable=True),
        24: Column(pa.String, nullable=True),
        25: Column(pa.String, nullable=True),
        26: Column(pa.String, nullable=True),
        27: Column(pa.String, nullable=True),
    }
)

auc_listings_schema = pa.DataFrameSchema(
    columns={
        "item": Column(pa.String),
        "quantity": Column(pa.Int),
        "buy": Column(pa.Int),
        "sellername": Column(pa.String),
        "price_per": Column(pa.Int),
        "time_remaining": Column(pa.Int),
    }
)

beancounter_data_raw_schema = pa.DataFrameSchema(
    columns={
        0: Column(pa.String, nullable=True),
        1: Column(pa.String, nullable=True),
        2: Column(pa.String, nullable=True),
        3: Column(pa.String, nullable=True),
        4: Column(pa.String, nullable=True),
        5: Column(pa.String, nullable=True),
        6: Column(pa.String, nullable=True),
        7: Column(pa.String, nullable=True),
        8: Column(pa.String, nullable=True),
        9: Column(pa.String, nullable=True),
        10: Column(pa.String, nullable=True),
        11: Column(pa.String, nullable=True),
        12: Column(pa.String, nullable=True),
    }
)

beancounter_purchases_schema = pa.DataFrameSchema(
    columns={
        "auction_type": Column(pa.String),
        "item": Column(pa.String),
        "buyer": Column(pa.String),
        "qty": Column(pa.Int),
        "buyout": Column(pa.Int, nullable=True),
        "bid": Column(pa.Int),
        "seller": Column(pa.String),
        "timestamp": Column(pa.DateTime),
        "buyout_per": Column(pa.Float),
        "bid_per": Column(pa.Float),
    }
)

beancounter_posted_schema = pa.DataFrameSchema(
    columns={
        "auction_type": Column(pa.String),
        "item": Column(pa.String),
        "seller": Column(pa.String),
        "qty": Column(pa.Int),
        "deposit": Column(pa.Int),
        "buyout": Column(pa.Int, nullable=True),
        "bid": Column(pa.Int),
        "timestamp": Column(pa.DateTime),
        "buyout_per": Column(pa.Float),
        "bid_per": Column(pa.Float),
    }
)

beancounter_failed_schema = pa.DataFrameSchema(
    columns={
        "auction_type": Column(pa.String),
        "item": Column(pa.String),
        "seller": Column(pa.String),
        "qty": Column(pa.Int),
        "deposit": Column(pa.Int),
        "buyout": Column(pa.Int, nullable=True),
        "bid": Column(pa.Int),
        "timestamp": Column(pa.DateTime),
        "buyout_per": Column(pa.Float),
        "bid_per": Column(pa.Float),
    }
)

beancounter_success_schema = pa.DataFrameSchema(
    columns={
        "auction_type": Column(pa.String),
        "item": Column(pa.String),
        "seller": Column(pa.String),
        "qty": Column(pa.Int),
        "received": Column(pa.Int),
        "deposit": Column(pa.Int),
        "ah_cut": Column(pa.Int),
        "buyout": Column(pa.Int, nullable=True),
        "bid": Column(pa.Int),
        "buyer": Column(pa.String),
        "timestamp": Column(pa.DateTime),
        "received_per": Column(pa.Float),
        "buyout_per": Column(pa.Float),
        "bid_per": Column(pa.Float),
    }
)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pricer']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML==5.2',
 'SLPP==1.2',
 'beautifulsoup4>=4.9.1,<5.0.0',
 'click==6.7',
 'deepdiff>=5.0.2,<6.0.0',
 'fastparquet==0.3.3',
 'importlib_metadata>=1.7.0,<2.0.0',
 'pandas==1.0.1',
 'pandera>=0.4.5,<0.5.0',
 'seaborn==0.9.0',
 'selenium>=3.141.0,<4.0.0',
 'sklearn>=0.0,<0.1',
 'tqdm>=4.49.0,<5.0.0']

entry_points = \
{'console_scripts': ['pricer = pricer.run:main']}

setup_kwargs = {
    'name': 'pricer',
    'version': '0.4.1',
    'description': 'Use WoW addon data to optimize auction buying and selling policies',
    'long_description': "# Pricer for WoW Auctions\n\n[![Tests](https://github.com/bluemania/wow_auctions/workflows/Tests/badge.svg)](https://github.com/bluemania/wow_auctions/actions?workflow=Tests)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Codecov](https://codecov.io/gh/bluemania/wow_auctions/branch/master/graph/badge.svg)](https://codecov.io/gh/bluemania/wow_auctions)\n[![PyPI](https://img.shields.io/pypi/v/pricer.svg)](https://pypi.org/project/pricer/)\n[![Documentation Status](https://readthedocs.org/projects/pricer/badge/?version=latest)](https://pricer.readthedocs.io/en/latest/?badge=latest)\n\n## About the project\n\nPricer for WoW Auction is a command line tool to helps users automate the decision making required to make profit on the World of Warcraft classic auction house.\nIt does so by reading addon data, using historical pricing information from [booty bay](bootybaygazette.com), and user specified preference, to calculate policies.\nThese policies take all available information to produce optimal actions for items, i.e. buying volume, selling low, selling high and crafting.\nThe policies overwrite the addon data, so that upon re-entering the game, the interface is ready to enact the optimal actions.\n\nThis is primarily a hobby project.\nMy aim is to learn about good software development practices, and apply my data science skills to an interesting problem.\nAnd the extra gold makes raiding even more fun!\n\nI did some [twitch streaming](https://www.twitch.tv/bluemania2) for a while, where I discussed the project goals and showed how things worked in more detail.\n\n## Built with\n\nMajor frameworks include:\n\n* Python 3+\n* Chromedriver and selenium - for booty bay data\n* Nox - for CI/CD process\n* Poetry - For dependencies, virtual environment, versioning and packaging\n\nRequirements for World of Warcraft classic include:\n\n* An active World of Warcraft account\n* An active Booty Bay Gazette subscription\n* ArkInventory addon - inventory tracking\n* Auctioneer addon - scanning auctions and enacting buy and sell policies\n* Beancounter addon - tracking player auction activity (comes with auctioneer)\n* Trade Skill Master - enacting crafting policies, also recommended to speed up moving items to bank and buying materials from vendors\n\n## Getting started\n\nThe following are instructions on how to set up the project locally.\n\nThe latest stable code can be found on [pypi](https://pypi.org/project/pricer/).\n\nDownload locally using:\n\n```bash\npip install pricer\n```\n\nFor development, clone the repository.\n\nThis project uses [poetry](https://python-poetry.org/) to manage dependencies, virtual environments, versioning and packaging.\n\n```bash\npoetry install\n```\n\n# Usage\n\n## User configuration\n\nIt is recommended to edit the `config/user_items.yaml` file, as this is the primary mechanism that users can select items of interest for buying, selling and crafting.\nRefer to the file to interpret the structure, ensure that items are named correctly.\n\nThe `config/user_settings.yaml` file should also be edited with information about paths to installations (WoW and Chromedriver), and active accounts.\nAccount names can be referenced from your WoW directory.\nBooty Bay information should be specified to be specific to your server.\nSpecify at least one of your characters as an `ahm` (auction house main), with others as `mule`.\n\nYou can optionally create a `SECRETS.yaml` file in the root directory with the following format.\nThis is useful to help automate the booty bay data feed, but not required.\nWe highly recommend using the Blizzard authenticator (stay safe!).\n\n```yaml\naccount: youraccount\npassword: yourpassword\n```\n\n## Running the program\n\nIt is recommended to perform a auction scan immediately prior to running the program.\nYou will need to be logged out while the program is running, as this is the only way the latest addon data can be loaded, and modified.\n\nRun the program using the following command from command line:\n\n```bash\npricer\n```\n\nAdditional flags can, and should be entered:\n\n* `-v` or `-vv` is useful for debugging purposes\n* `-b` is used to seek a refresh of Booty Bay data; it is recommended to seek an update at least once every day or so.\n* `-h` for help on additional flags and functionality that may be available\n\nIf the run has been successful, you should see some tabular information printed in the console.\nThis will include information about what items to make and expected profits for feasible selling items.\n\nYou should see Auctioneer data has been changed, so that (feasible) sell prices and thresholds for buying using snatch have been set.\n\n## Tests and tooling\n\nThis project seeks to use modern code quality and CI/CD tooling including\n\n* Dependency management, virtual env and packaging (poetry)\n* Linting (black, flake8, darglint)\n* Type checking (mypy)\n* Testing (pytest, pandera, codecov)\n* Docs (sphinx, autodoc)\n* Task automation CI/CD (nox, pre-commit, github actions, pr-labeler, release-drafter)\n* Publishing (pypi, readthedocs)\n\n# Contributing\n\nThis project is pre-release and under development. \n\nUsers are welcome to try the program, fork, or [contribute](CONTRIBUTING.md), however [support](SUPPORT.md) is not guarenteed.\n\nFollow this link for instructions on managing [releases](RELEASE.md).\n\n# License\n\nAll assets and code are under the MIT LICENSE and in the public domain unless specified otherwise.\nSee the [license](LICENSE) for more info.\n\n# Contact\n\nFeel free to reach out in-game; you'll see me on Grobbulus on Amazona. \n\nYou can leave an open issue seeking to connect and I'll get back to you.\n\nI also occassionally stream project development on [twitch](https://www.twitch.tv/bluemania2).\n",
    'author': 'bluemania',
    'author_email': 'damnthatswack@hotmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/bluemania/wow_auctions',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
